# Use this node in the ast when you want to add some custom C++
#
# This is one mechanism to allow for a leaky abstraction.
import ast
from typing import Callable, Optional, cast

import func_adl_xAOD.common.statement as statements
from func_adl_xAOD.common.cpp_representation import cpp_value, cpp_variable
from func_adl_xAOD.common.util_scope import gc_scope

# The list of methods and the re-write functions for them. Each rewrite function
# is called with the Call node, which includes arguments, names, etc. It should return
# None or a cpp_ast.
method_names = {}


class CPPCodeValue (ast.AST):
    r'''
    Represents a C++ bit of code that returns a value. Like a function call or a member call.
    Use the be-fore the wire visit phase of processing to look for a pattern that needs
    to generate AST code, like a method call. Then place this AST in place of the function.
    The back-end will then do the rendering useing the information included below.
    '''

    def __init__(self):
        # Files that need to be included at the top of the generated C++ file
        self.include_files = []

        # Code that is run once at the start of each "event"
        self.initialization_code = []

        # Code that is run when the particular bit of code needs to be invoked (e.g. in the middle of a hot loop).
        # This is invoked in its own scope (between "{" and "}") so there are no variable collisions.
        self.running_code = []

        # The arguments to the function. These are "correctly" mapped into the argument values
        # that are passed to the function and then a text replacement is done in the code.
        self.args = []

        # Special replacement if this is a method call. A tuple. THe first item is the string to be replaced in the
        # code. The second is the name against which we should be making the call (e.g. if j is the current jet variable,
        # the tuple might be ("obj", "j")).
        self.replacement_instance_obj = None

        # A string representing the result value. This must be a simple variable. It will get replaced
        # in all the code lines above.
        self.result: Optional[str] = None

        # Representation to use for the resulting variable. Includes C++ type information.
        # A lambda that takes teh scope as an argument and returns a cpp variable to hold things.
        self.result_rep: Optional[Callable[[gc_scope], cpp_variable]] = None

        # We have no further fields for the ast machinery to explore, so this is empty for now.
        self.fields = []


class cpp_ast_finder(ast.NodeTransformer):
    r'''
    Look through the complete ast and replace method calls that are to a C++ plug in with a c++ ast
    node.
    '''
    def __init__(self, method_names: dict):
        self._method_names = method_names

    def try_call(self, name, node):
        'Try to use name to do the call. Returns (ok, result) monad'
        if name in self._method_names:
            cpp_call_ast = self._method_names[name](node)
            return (cpp_call_ast is not None, cpp_call_ast)
        return (False, None)

    def visit_Call(self, node):
        r'''
        Looking for a member call of a particular name. We rewrite that as
        another name.
        WARNING: currently the namespace is global, so the parent type doesn't matter!
        '''

        # Make sure all parts of this AST are visited properly before we attempt to
        # understand the call.
        self.generic_visit(node)

        # Examine the func to see if this is a member call.
        func = node.func
        if (type(func) is ast.Attribute) and (type(func.value) is ast.Name):
            ok, new_node = self.try_call(func.attr, node)
            if ok:
                return new_node
        elif type(func) is ast.Name:
            ok, new_node = self.try_call(func.id, node)
            if ok:
                return new_node

        return node


def process_ast_node(visitor, gc, call_node: ast.Call):
    r'''Inject the proper code into the output stream to deal with this C++ code.

    We expect this to be run on the back-end of the system.

    visitor - The node visitor that is converting the code into C++
    gc - the generated code object that we fill with actual code
    call_node - a Call ast node, with func being a CPPCodeValue.

    Result:
    representation - A value that represents the output
    '''

    # We write everything into a new scope to prevent conflicts. So we have to declare the result ahead of time.
    cpp_ast_node = cast(CPPCodeValue, call_node.func)
    result_rep = cpp_ast_node.result_rep(gc.current_scope())

    gc.declare_variable(result_rep)

    # Include files
    for i in cpp_ast_node.include_files:
        gc.add_include(i)

    # Build the dictionary for replacement for the object we are calling
    # against, if any.
    repl_list = []
    if cpp_ast_node.replacement_instance_obj is not None:
        repl_list += [(cpp_ast_node.replacement_instance_obj[0], visitor.resolve_id(cpp_ast_node.replacement_instance_obj[1]).rep.as_cpp())]

    # Process the arguments that are getting passed to the function
    for arg, dest in zip(cpp_ast_node.args, call_node.args):
        rep = visitor.get_rep(dest)
        repl_list += [(arg, rep.as_cpp())]

    # Emit the statements.
    blk = statements.block()
    visitor._gc.add_statement(blk)

    for s in cpp_ast_node.running_code:
        l_s = s
        for src, dest in repl_list:
            l_s = l_s.replace(src, str(dest))
        blk.add_statement(statements.arbitrary_statement(l_s))

    # Set the result and close the scope
    assert cpp_ast_node.result is not None
    blk.add_statement(statements.set_var(result_rep, cpp_value(cpp_ast_node.result, gc.current_scope(), result_rep.cpp_type())))
    gc.pop_scope()

    return result_rep



# Drive the translate of the AST from start into a set of files, which one can then do whatever
# is needed to.
import ast
import os
import sys
from abc import ABC, abstractmethod
from collections import namedtuple
from pathlib import Path

import func_adl_xAOD.common.cpp_ast as cpp_ast
import func_adl_xAOD.common.cpp_representation as crep
import jinja2
from func_adl.ast.aggregate_shortcuts import aggregate_node_transformer
from func_adl.ast.func_adl_ast_utils import change_extension_functions_to_calls
from func_adl.ast.function_simplifier import simplify_chained_calls
from func_adl_xAOD.common.ast_to_cpp_translator import query_ast_visitor
from func_adl_xAOD.common.cpp_functions import find_known_functions
from func_adl_xAOD.common.util_scope import top_level_scope

ExecutionInfo = namedtuple('ExecutionInfo', 'result_rep output_path main_script all_filenames')


class _cpp_source_emitter:
    r'''
    Helper class to emit C++ code as we go
    '''

    def __init__(self):
        self._lines_of_query_code = []
        self._indent_level = 0

    def add_line(self, ll):
        'Add a line of code, automatically deal with the indent'
        if ll == '}':
            self._indent_level -= 1

        self._lines_of_query_code += [
            "{0}{1}".format("  " * self._indent_level, ll)]

        if ll == '{':
            self._indent_level += 1

    def lines_of_query_code(self):
        return self._lines_of_query_code


# The following was copied from: https://www.oreilly.com/library/view/python-cookbook/0596001673/ch04s22.html
def _find(pathname: str, matchFunc=os.path.isfile):
    assert len(pathname) > 0
    for dirname in (sys.path + ['/usr/local']):
        candidate = os.path.join(dirname, pathname)
        if matchFunc(candidate):
            return candidate
    all_dirs = ','.join(sys.path + ['/usr/local'])
    raise RuntimeError(f"Can't find file '{pathname}'. Looked in {all_dirs}")


def _find_dir(path):
    return _find(path, matchFunc=os.path.isdir)


def _is_format_request(a: ast.AST) -> bool:
    '''Return true if the top level ast is a call to generate a ROOT file output.

    Args:
        ast (ast.AST): AST to check at the top level

    Returns:
        bool: True if the ast is not format agnostic.
    '''
    if not isinstance(a, ast.Call):
        raise ValueError(f'A func_adl ast must start with a function call. This does not: {ast.dump(a)}')
    if not isinstance(a.func, ast.Name):
        raise ValueError(f'A func_adl ast must start with a function call to something like Select or AsROOTTTree. This does not: {ast.dump(a)}')
    return a.func.id == 'ResultTTree'


class executor(ABC):
    def __init__(self, file_names: list, runner_name: str, template_dir_name: str, method_names: dict):
        self._file_names = file_names
        self._runner_name = runner_name
        self._template_dir_name = template_dir_name
        self._method_names = method_names

    def _copy_template_file(self, j2_env, info, template_file, final_dir: Path):
        'Copy a file to a final directory'
        j2_env.get_template(template_file).stream(info).dump(str(final_dir / template_file))

    def apply_ast_transformations(self, a: ast.AST):
        r'''
        Run through all the transformations that we have on tap to be run on the client side.
        Return a (possibly) modified ast.
        '''

        # Do tuple resolutions. This might eliminate a whole bunch fo code!
        a = change_extension_functions_to_calls(a)
        a = aggregate_node_transformer().visit(a)
        a = simplify_chained_calls().visit(a)
        a = find_known_functions().visit(a)

        # Any C++ custom code needs to be threaded into the ast
        a = cpp_ast.cpp_ast_finder(self._method_names).visit(a)

        # And return the modified ast
        return a

    @abstractmethod
    def get_visitor_obj(self) -> query_ast_visitor:
        pass

    def write_cpp_files(self, ast: ast.AST, output_path: Path) -> ExecutionInfo:
        r"""
        Given the AST generate the C++ files that need to run. Return them along with
        the input files.
        """

        # Find the base file dataset and mark it.
        from func_adl import find_EventDataset
        file = find_EventDataset(ast)
        iterator = crep.cpp_variable("bogus-do-not-use", top_level_scope(), cpp_type=None)
        file.rep = crep.cpp_sequence(iterator, iterator, top_level_scope())  # type: ignore

        # Visit the AST to generate the code structure and find out what the
        # result is going to be.
        qv = self.get_visitor_obj()
        result_rep = qv.get_rep(ast) if _is_format_request(ast) \
            else qv.get_as_ROOT(ast)

        # Emit the C++ code into our dictionaries to be used in template generation below.
        query_code = _cpp_source_emitter()
        qv.emit_query(query_code)
        book_code = _cpp_source_emitter()
        qv.emit_book(book_code)
        class_decl_code = qv.class_declaration_code()
        includes = qv.include_files()

        # The replacement dict to pass to the template generator can now be filled
        info = {}
        info['query_code'] = query_code.lines_of_query_code()
        info['class_decl'] = class_decl_code
        info['book_code'] = book_code.lines_of_query_code()
        info['include_files'] = includes

        # We use jinja2 templates. Write out everything.
        template_dir = _find_dir(self._template_dir_name)
        j2_env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(template_dir))

        for file_name in self._file_names:
            self._copy_template_file(j2_env, info, file_name, output_path)

        (output_path / self._runner_name).chmod(0o755)

        # Build the return object.
        return ExecutionInfo(result_rep, output_path, self._runner_name, self._file_names)



import func_adl_xAOD.common.cpp_types as ctyp
from func_adl_xAOD.common.event_collections import (
    event_collection_collection, event_collection_container, event_collections)


class atlas_xaod_event_collection_container(event_collection_container):
    def __init__(self, type_name, is_pointer=True):
        super().__init__(type_name, is_pointer)

    def __str__(self):
        return f"edm::Handle<{self._type_name}>"


class atlas_xaod_event_collection_collection(event_collection_collection):
    def __init__(self, type_name, element_name, is_type_pointer=True, is_element_pointer=True):
        super().__init__(type_name, element_name, is_type_pointer, is_element_pointer)

    def __str__(self):
        return f"const {self._type_name}*"


# all the collections types that are available. This is required because C++
# is strongly typed, and thus we have to transmit this information.
atlas_xaod_collections = [
    {
        'function_name': "Jets",
        'include_files': ['xAODJet/JetContainer.h'],
        'container_type': atlas_xaod_event_collection_collection('xAOD::JetContainer', 'xAOD::Jet'),
    },
    {
        'function_name': "Tracks",
        'include_files': ['xAODTracking/TrackParticleContainer.h'],
        'container_type': atlas_xaod_event_collection_collection('xAOD::TrackParticleContainer', 'xAOD::TrackParticle')
    },
    {
        'function_name': "EventInfo",
        'include_files': ['xAODEventInfo/EventInfo.h'],
        'container_type': atlas_xaod_event_collection_container('xAOD::EventInfo'),
        'is_collection': False,
    },
    {
        'function_name': "TruthParticles",
        'include_files': ['xAODTruth/TruthParticleContainer.h', 'xAODTruth/TruthParticle.h', 'xAODTruth/TruthVertex.h'],
        'container_type': atlas_xaod_event_collection_collection('xAOD::TruthParticleContainer', 'xAOD::TruthParticle')
    },
    {
        'function_name': "Electrons",
        'include_files': ['xAODEgamma/ElectronContainer.h', 'xAODEgamma/Electron.h'],
        'container_type': atlas_xaod_event_collection_collection('xAOD::ElectronContainer', 'xAOD::Electron')
    },
    {
        'function_name': "Muons",
        'include_files': ['xAODMuon/MuonContainer.h', 'xAODMuon/Muon.h'],
        'container_type': atlas_xaod_event_collection_collection('xAOD::MuonContainer', 'xAOD::Muon')
    },
    {
        'function_name': "MissingET",
        'include_files': ['xAODMissingET/MissingETContainer.h', 'xAODMissingET/MissingET.h'],
        'container_type': atlas_xaod_event_collection_collection('xAOD::MissingETContainer', 'xAOD::MissingET'),
    },
]


class atlas_xaod_event_collections(event_collections):
    def __init__(self):
        super().__init__(atlas_xaod_collections)

    def get_running_code(self, container_type: event_collection_container) -> list:
        return [f'{container_type} result = 0;',
                'ANA_CHECK (evtStore()->retrieve(result, collection_name));']


# Configure some info about the types.
ctyp.add_method_type_info("xAOD::TruthParticle", "prodVtx", ctyp.terminal('xAODTruth::TruthVertex', is_pointer=True))
ctyp.add_method_type_info("xAOD::TruthParticle", "decayVtx", ctyp.terminal('xAODTruth::TruthVertex', is_pointer=True))
ctyp.add_method_type_info("xAOD::TruthParticle", "parent", ctyp.terminal('xAOD::TruthParticle', is_pointer=True))
ctyp.add_method_type_info("xAOD::TruthParticle", "child", ctyp.terminal('xAOD::TruthParticle', is_pointer=True))



# Some math utilities
import func_adl_xAOD.common.cpp_ast as cpp_ast
from func_adl_xAOD.common.cpp_representation import cpp_variable
from func_adl_xAOD.common.cpp_vars import unique_name


def DeltaRAst(call_node):
    r'''
    User is trying to call DeltaR (eta1, phi1, eta2, phi2). We turn this into a call
    into a call into ROOT that does the phi subtraction (I can never get that crap right).
    '''

    if len(call_node.args) != 4:
        raise ValueError("Calling DeltaR(eta1, phi1, eta2, phi2) has incorrect number of arguments")

    # Create an AST to hold onto all of this.
    r = cpp_ast.CPPCodeValue()
    # We need TVector2 included here
    r.include_files += ['TVector2.h', 'math.h']

    # We need all four arguments pushed through.
    r.args = ['eta1', 'phi1', 'eta2', 'phi2']

    # The code is three steps
    r.running_code += ['auto d_eta = eta1 - eta2;']
    r.running_code += ['auto d_phi = TVector2::Phi_mpi_pi(phi1-phi2);']
    r.running_code += ['auto result = sqrt(d_eta*d_eta + d_phi*d_phi);']
    r.result = 'result'
    r.result_rep = lambda scope: cpp_variable(unique_name('delta_r'), scope=scope, cpp_type='double')

    call_node.func = r
    return call_node


def DeltaR(eta1, phi1, eta2, phi2):
    'Calculate the DeltaR between two eta,phi specified vectors'
    raise NotImplementedError('DeltaR should never be called in python!')


def get_math_methods():
    return {
        'DeltaR': DeltaRAst
    }



#!/usr/bin/env python

import FWCore.ParameterSet.Config as cms  # type: ignore

process = cms.Process("Demo")

process.load("FWCore.MessageService.MessageLogger_cfi")

process.maxEvents = cms.untracked.PSet(input=cms.untracked.int32(-1))

filelistPath = 'filelist.txt'
fileNames = tuple(['file:{0}'.format(line) for line in open(filelistPath, 'r').readlines()])

process.source = cms.Source("PoolSource",
                            # replace 'myfile.root' with the source file you want to use
                            fileNames=cms.untracked.vstring(
                                *fileNames
                            )
                            )

process.demo = cms.EDAnalyzer('Analyzer'
                              )

process.TFileService = cms.Service("TFileService",
                                   fileName=cms.string('/results/ANALYSIS.root')
                                   )

process.p = cms.Path(process.demo)



# Simple type system to help reason about types as they go through the system.


class terminal:
    'Represents something we cannot see inside, like float, or int, or bool'

    def __init__(self, t: str, is_pointer: bool = False):
        '''Create a terminal type - a type that we do not need to see inside

        Args:
            t (str): The name of the type (integer, double, xAOD::Jet, etc.)
            is_pointer (bool, optional): True if this is a pointer. Defaults to False.
        '''
        self._type = t
        self._is_pointer = is_pointer

    def __str__(self):
        return str(self.type)

    def is_pointer(self):
        return self._is_pointer

    def default_value(self):
        raise NotImplementedError()
        # if self._type == "double":
        #     return "0.0"
        # elif self._type == "float":
        #     return "0.0"
        # elif self._type == "int":
        #     return "0"
        # else:
        #     raise Exception(f"Do not know a default value for the type '{self._type}'.")

    @property
    def type(self) -> str:
        return self._type

    def get_dereferenced_type(self):
        return terminal(self._type, is_pointer=False) if self._is_pointer else self


class collection (terminal):
    'Represents a collection/list/vector of the same type'

    def __init__(self, t, is_pointer=False):
        '''
        Initialize a collection type.

        t:      The type of each element in the collection
        '''
        array_type = f"std::vector<{t}>"
        terminal.__init__(self, array_type, is_pointer)
        self._element_type = t

    def element_type(self):
        return self._element_type


class tuple:
    'Represents a value which is a collection of other types'

    def __init__(self, type_list):
        '''
        Initialize a type list. The value consists of `len(type_list)` items, each
        of the type held inside type_lits.

        type_list:      tuple,etc., that we can iterate over to get the types.
        '''
        self._type_list = type_list

    def __str__(self):
        return "(" + ','.join(self._type_list) + ")"

###########################
# Manage types


g_method_type_dict = {}


def add_method_type_info(type_string, method_name, t):
    '''
    Define a return type for a method

    type_string         String of the object the method is calling against
    method_name         Name of the object
    t                   The type (terminal, collection, etc.) of return type
    '''
    if type_string not in g_method_type_dict:
        g_method_type_dict[type_string] = {}
    g_method_type_dict[type_string][method_name] = t


def method_type_info(type_string, method_name):
    '''
    Return the type of the method's return value
    '''
    if type_string not in g_method_type_dict:
        return None
    if method_name not in g_method_type_dict[type_string]:
        return None
    return g_method_type_dict[type_string][method_name]



from func_adl_xAOD.common.ast_to_cpp_translator import query_ast_visitor
from func_adl_xAOD.common.statement import book_ttree, ttree_fill


class book_xaod_ttree(book_ttree):
    'Book an ATLAS TTree for writing out. Meant to be in the Book method'

    def __init__(self, tree_name, leaves):
        super().__init__(tree_name, leaves)

    def emit(self, e):
        'Emit the book statement for a tree'
        e.add_line('ANA_CHECK (book (TTree ("{0}", "My analysis ntuple")));'.format(
            self._tree_name))
        e.add_line('auto myTree = tree ("{0}");'.format(self._tree_name))
        for var_pair in self._leaves:
            e.add_line('myTree->Branch("{0}", &{1});'.format(var_pair[0], var_pair[1].as_cpp()))


class xaod_ttree_fill(ttree_fill):
    'Fill a ATLAS TTree'

    def __init__(self, tree_name):
        super().__init__(tree_name)

    def emit(self, e):
        e.add_line('tree("{0}")->Fill();'.format(self._tree_name))


class atlas_xaod_query_ast_visitor(query_ast_visitor):
    r"""
    Drive the conversion to C++ from the top level query
    for ATLAS xAOD
    """

    def __init__(self):
        prefix = 'atlas_xaod'
        is_loop_var_a_ref = False
        super().__init__(prefix, is_loop_var_a_ref)

    def create_book_ttree_obj(self, tree_name: str, leaves: list) -> book_ttree:
        return book_xaod_ttree(tree_name, leaves)

    def create_ttree_fill_obj(self, tree_name: str) -> ttree_fill:
        return xaod_ttree_fill(tree_name)



from func_adl_xAOD.atlas.xaod.event_collections import \
    atlas_xaod_event_collections as ec
from func_adl_xAOD.atlas.xaod.jets import get_jet_methods
from func_adl_xAOD.atlas.xaod.query_ast_visitor import \
    atlas_xaod_query_ast_visitor
from func_adl_xAOD.common.executor import executor


class atlas_xaod_executor(executor):
    def __init__(self):
        file_names = ['ATestRun_eljob.py', 'package_CMakeLists.txt', 'query.cxx', 'query.h', 'runner.sh']
        runner_name = 'runner.sh'
        template_dir_name = 'func_adl_xAOD/template/atlas/r21'
        method_names = ec().get_method_names()
        method_names.update(get_jet_methods())
        super().__init__(file_names, runner_name, template_dir_name, method_names)

    def get_visitor_obj(self):
        return atlas_xaod_query_ast_visitor()



import func_adl_xAOD.common.cpp_types as ctyp
from func_adl_xAOD.common.cpp_representation import cpp_value
from func_adl_xAOD.common.cpp_vars import unique_name


##################
# TTree return
class cpp_ttree_rep(cpp_value):
    'This is what a TTree operator returns'

    def __init__(self, filename, treename, scope):
        cpp_value.__init__(self, unique_name("ttree_rep"), scope, ctyp.terminal("ttreetfile"))
        self.filename = filename
        self.treename = treename



# Code to translate from a reduced AST into C++ code. This is done by traversing the
# Python AST code.

import ast
import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Type, Union, cast

import func_adl_xAOD.common.cpp_ast as cpp_ast
import func_adl_xAOD.common.cpp_representation as crep
import func_adl_xAOD.common.cpp_types as ctyp
import func_adl_xAOD.common.math_utils  # NOQA
import func_adl_xAOD.common.result_ttree as rh
import func_adl_xAOD.common.statement as statement
from func_adl.ast.call_stack import argument_stack, stack_frame
from func_adl.ast.func_adl_ast_utils import FuncADLNodeVisitor, function_call
from func_adl.util_ast import lambda_unwrap
from func_adl_xAOD.common.cpp_functions import FunctionAST
from func_adl_xAOD.common.cpp_vars import unique_name
from func_adl_xAOD.common.generated_code import generated_code
from func_adl_xAOD.common.util_scope import (deepest_scope, gc_scope,
                                             gc_scope_top_level,
                                             top_level_scope)
from func_adl_xAOD.common.utils import most_accurate_type

# Convert between Python comparisons and C++.
compare_operations = {
    ast.Lt: '<',
    ast.LtE: '<=',
    ast.Gt: '>',
    ast.GtE: '>=',
    ast.Eq: '==',
    ast.NotEq: '!=',
}


# Unary operators - we aren't doing not and invert just yet.
_known_unary_operators: Dict[Type, str] = {
    ast.UAdd: '+',
    ast.USub: '-',
    ast.Not: '!',
}


# Known binary operators
_known_binary_operators: Dict[Type, str] = {
    ast.Add: '+',
    ast.Sub: '-',
    ast.Mult: '*',
    ast.Div: '/',
    ast.Mod: '%',
}


class xAODTranslationError(Exception):
    'Thrown when a translation error happens of one sort or another.'

    def __init__(self, msg):
        Exception.__init__(self, msg)


def check_accumulator_type(t: ctyp.terminal):
    'We can only deal with certain types for doing an accumulation. Make sure this is one.'
    t_str = str(t)
    return (t_str == "float") or (t_str == "double") or (t_str == "int")


def guess_type_from_number(n):
    if int(n) == n:
        return ctyp.terminal("int")
    return ctyp.terminal("double")


def rep_is_collection(rep) -> bool:
    if isinstance(rep, crep.cpp_sequence):
        return True
    if isinstance(rep, crep.cpp_collection):
        return True
    return False


def get_ttree_type(rep):
    'Looking at a rep, figure out how it should get stored in a tree'
    if isinstance(rep, crep.cpp_sequence):
        if not isinstance(rep.sequence_value(), (crep.cpp_value, crep.cpp_sequence)):
            raise Exception("Nested data structures (2D arrays, etc.) in TTree's are not yet supported. Numbers or arrays of numbers only for now.")
        return ctyp.collection(rep.sequence_value().cpp_type())
    else:
        return rep.cpp_type()


def determine_type_mf(parent_type, function_name):
    '''
    Determine the return type of the member function. Do our best to make
    an intelligent case when we can.

    parent_type:        the type of the parent
    function_name:      the name of the function we are calling
    '''
    # If we don't know the type...
    if parent_type is None:
        raise RuntimeError("Internal Error: Trying to call member function for a type we do not know!")
    # If we are doing one of the normal "terminals", then we can just bomb. This should not happen!

    rtn_type = ctyp.method_type_info(str(parent_type), function_name)
    if rtn_type is not None:
        return rtn_type

    # We didn't know it. Lets make a guess, and error out if we are clearly making a mistake.
    base_types = ['double', 'float', 'int']
    s_parent_type = str(parent_type)
    if s_parent_type in base_types:
        raise xAODTranslationError(f'Unable to call method {function_name} on type {str(parent_type)}.')

    # Ok - we give up. Return a double.
    logging.getLogger(__name__).warning(f"Warning: assumping that the method '{str(s_parent_type)}.{function_name}(...)' has return type 'double'. Use cpp_types.add_method_type_info to suppress (or correct) this warning.")
    return ctyp.terminal('double')


def _extract_column_names(names_ast: ast.AST) -> List[str]:
    'Extract a list of strings from an ast using literal evaluation. A single name is returned as a list.'
    names = ast.literal_eval(names_ast)
    if isinstance(names, str):
        return [names]
    return names


class query_ast_visitor(FuncADLNodeVisitor, ABC):
    r"""
    Drive the conversion to C++ from the top level query
    """

    def __init__(self, prefix, is_loop_var_a_ref):
        r'''
        Initialize the visitor.
        '''
        # Tracks the output of the code.
        self._gc = generated_code()
        self._arg_stack = argument_stack()
        self._result = None
        self._prefix = prefix
        self._is_loop_var_a_ref = is_loop_var_a_ref

    def include_files(self):
        return self._gc.include_files()

    def emit_query(self, e):
        'Emit the parsed lines'
        self._gc.emit_query_code(e)

    def emit_book(self, e):
        'Emit the parsed lines'
        self._gc.emit_book_code(e)

    def class_declaration_code(self):
        return self._gc.class_declaration_code()

    def visit(self, node):
        '''Visit a node. If the node already has a rep, then it has been visited and we
        do not need to visit it again.

        node - if the node has a rep, just return

        '''
        if not hasattr(node, 'rep'):
            FuncADLNodeVisitor.visit(self, node)

        # Lots of nodes never get a representation - like `ast.Name`, so we
        # need to protect the access here.
        if hasattr(node, 'rep'):
            self._result = node.rep

    def get_rep(self, node: ast.AST, retain_scope: bool = False) -> Union[crep.cpp_value, crep.cpp_sequence]:
        r'''Return the rep for the node. If it isn't set yet, then run our visit on it.

        node - The ast node to generate a representation for.
        retain_scope - If true, then the scope level will remain the same before and after the call.

        RULES for .rep: NEVER access it without using get_rep. It is fine, of course, if you are setting it as a result
        of visiting. BUT ALWAYS GO THROUGH get_rep to get the rep for a node you aren't handling directly. If you ever find yourself
        writing a "hasattr(node, 'rep')" you will almost certainly be introducing a bug. Use get_rep instead!!
        '''
        # If the rep is present, make sure it is still valid by checking the scope.
        result = None
        if hasattr(node, 'rep'):
            result = node.rep  # type: ignore
            if not self._gc.current_scope().starts_with(result.scope()):
                result = None

        # If this node already has a representation, then it has been
        # processed and we do not need to do it again.
        if result is None:
            s = self._gc.current_scope() if retain_scope else None
            self.visit(node)
            if s is not None:
                self._gc.set_scope(s)

        # If it still didn't work, this is an internal error. But make the error message a bit nicer.
        if not hasattr(node, 'rep'):
            raise Exception('Internal Error: attempted to get C++ representation for AST node "{0}", but failed.'.format(ast.dump(node)))
        self._result = node.rep  # type: ignore

        return node.rep  # type: ignore

    @abstractmethod
    def create_ttree_fill_obj(self, tree_name: str) -> statement.ttree_fill:
        pass

    @abstractmethod
    def create_book_ttree_obj(self, tree_name: str, leaves: list) -> statement.book_ttree:
        pass

    def get_as_ROOT(self, node: ast.AST) -> rh.cpp_ttree_rep:
        '''For a given node, return a root ttree rep.

        This is used to make sure whatever the sequence is, that it returns a RootTTree. Use this
        when the user may not have explicitly requested an output format. For example, if they end
        with a tuple or a dict request, but not a AsAwkwardArray or similar.

        1. If the top level ast node already requests an ROOTTTree, the representation is returned.
        1. If the node is a sequence of `cpp_values`, then a single column tree is created.
        1. If the node is a sequence of dict's, a ttree is created that uses the dictionary
           key's as column names.
        1. Anything else causes an exception to be raised.

        Args:
            node (ast.AST): Top level `func_adl` expression that is to be renered as a ROOT file.

        Returns:
            rh.cpp_ttree_rep: The resulting node that will generate a root file.

        Exceptions:
            ValueError: If we end with something other than above, we raise `ValueError` to
                        indicate that we can't figure out how to convert something into a ROOT
                        file.
        '''
        r = self.get_rep(node)
        if isinstance(r, rh.cpp_ttree_rep):
            return r

        # If this isn't a sequence, then we are totally blown here.
        if not isinstance(r, crep.cpp_sequence):
            raise ValueError(f'Do not know how to convert {r} into a ROOT file')

        # If this is a dict, then pull out each item and re-assemble into a tuple
        # which we can feed to the root guy.
        values = r.sequence_value()
        if isinstance(values, crep.cpp_dict):
            values = cast(crep.cpp_dict, values)
            col_values = values.value_dict.values()
            col_names = ast.List(elts=list(values.value_dict.keys()))
            s_tuple = crep.cpp_tuple(tuple(col_values), values.scope())
            tuple_sequence = crep.cpp_sequence(s_tuple, r.iterator_value(), r.scope())  # type: ignore
            ast_dummy_source = ast.Constant(value=1)
            ast_dummy_source.rep = tuple_sequence  # type: ignore
            ast_ttree = function_call('ResultTTree',
                                      [ast_dummy_source,
                                       col_names,
                                       ast.parse(f'"{self._prefix}_tree"').body[0].value,  # type: ignore
                                       ast.parse(f'"{self._prefix}_output"').body[0].value])  # type: ignore
            result = self.get_rep(ast_ttree)
            assert isinstance(result, rh.cpp_ttree_rep)
            return result

        if isinstance(values, crep.cpp_tuple):
            col_names = ast.List(elts=[ast.parse(f"'col{i}'").body[0].value for i, _ in enumerate(values.values())])  # type: ignore
            ast_ttree = function_call('ResultTTree',
                                      [node,
                                       col_names,
                                       ast.parse(f'"{self._prefix}_tree"').body[0].value,  # type: ignore
                                       ast.parse(f'"{self._prefix}_output"').body[0].value])  # type: ignore
            result = self.get_rep(ast_ttree)
            assert isinstance(result, rh.cpp_ttree_rep)
            return result
        elif isinstance(values, crep.cpp_value) or isinstance(values, crep.cpp_sequence):
            ast_ttree = function_call('ResultTTree',
                                      [node,
                                       ast.parse('"col1"').body[0].value,  # type: ignore
                                       ast.parse(f'"{self._prefix}_tree"').body[0].value,  # type: ignore
                                       ast.parse(f'"{self._prefix}_output"').body[0].value])  # type: ignore
            result = self.get_rep(ast_ttree)
            assert isinstance(result, rh.cpp_ttree_rep)
            return result
        else:
            raise ValueError(f'Do not know how to convert a sequence of {r.sequence_value()} into a ROOT file.')

        raise NotImplementedError()

    def get_rep_value(self, node, retain_scope=False) -> crep.cpp_value:
        r'''Return the rep for the node. If it isn't set yet, then run our visit on it. Assure we are returning a value

        node - The ast node to generate a representation for.
        retain_scope - If true, then the scope level will remain the same before and after the call.
        '''
        v = self.get_rep(node, retain_scope)
        if not isinstance(v, crep.cpp_value):
            raise Exception("Expected a cpp value! Internal error")
        return v

    def resolve_id(self, id):
        'Look up the in our local dict. This takes care of function arguments, etc.'
        return self._arg_stack.lookup_name(id)

    def make_sequence_from_collection(self, rep):
        '''
        Take a collection and produce a sequence. Eventually this should likely be some sort of
        plug-in architecture. But for now, we will just assume everything looks like a vector. When
        it comes time for a new type, this is where it should go.
        '''
        element_type = rep.cpp_type().element_type()
        element_type = element_type.get_dereferenced_type() if self._is_loop_var_a_ref else element_type
        iterator_value = crep.cpp_value(unique_name("i_obj"), None, element_type)  # type: ignore
        l_statement = statement.loop(iterator_value, crep.dereference_var(rep), is_loop_var_a_ref=self._is_loop_var_a_ref)  # type: ignore
        self._gc.add_statement(l_statement)
        iterator_value.reset_scope(self._gc.current_scope())

        # For a new sequence like this the sequence and iterator value are the same
        return crep.cpp_sequence(iterator_value, iterator_value, self._gc.current_scope())

    def as_sequence(self, generation_ast: ast.AST):
        r'''
        We will convert the generation_ast into a sequence if we can. If we can't, that indicates
        a likely programming error by this library or by the user.

        generation_ast - The AST that will generate the collection (a call to something that
                         returns a collection or a Select statement, etc.)

        returns:

        sequence:       An object of type crep.cpp_sequence that contains all the information
                        about the sequence.
        '''
        # Get the representation for the ast
        rep = self.get_rep(generation_ast)

        # If this is already a sequence then we are done!
        if isinstance(rep, crep.cpp_sequence):
            return rep

        # Next do a lookup to see if we have already defined a sequence at this level
        r = self._gc.get_rep(rep)
        if r is not None:
            return r

        # If this is a collection, then we need to turn it into a sequence.
        if isinstance(rep, crep.cpp_collection):
            r = self.make_sequence_from_collection(rep)
            self._gc.set_rep(rep, r)
            return r

        # If it isn't a sequence or a collection, then something has gone wrong.
        raise Exception(f"Unable to generate a sequence from the given AST. Either there is an internal error, or you are trying to manipulate a '{type(rep).__name__}' as a sequence (ast is: {ast.dump(generation_ast)})")

    def visit_Call_Lambda(self, call_node):
        'Call to a lambda function. We propagate the arguments through the function'

        with stack_frame(self._arg_stack):
            for c_arg, l_arg in zip(call_node.args, call_node.func.args.args):
                self._arg_stack.define_name(l_arg.arg, c_arg)

            # Next, process the lambda's body.
            call_node.rep = self.get_rep(call_node.func.body)

    def _create_accumulator(self, seq: crep.cpp_sequence, acc_type: ctyp.terminal, initial_value=None):
        'Helper to create an accumulator for the Aggregate function'
        accumulator_type = acc_type

        # When we implement other types of aggregate, this code will need to
        # be back in.
        # if accumulator_type is None:
        #     sv = seq.sequence_value()
        #     if not isinstance(sv, crep.cpp_value):
        #         raise Exception("Do not know how to accumulate a sequence!")
        #     accumulator_type = sv.cpp_type()
        if not check_accumulator_type(accumulator_type):
            raise ValueError(f"Aggregate over a sequence of type '{str(accumulator_type)}' is not supported.")

        # Getting the scope level right is tricky. If this is a straight sequence of items, then we want the sequence level.
        # But if this is a sequence of sequences, we are aggregating over the sequence itself. So we need to do it one level
        # up from where the iterator is running on the interior sequence.
        seq_val = seq.sequence_value()
        if isinstance(seq_val, crep.cpp_sequence):
            accumulator_scope = seq_val.iterator_value().scope()[-1]
        else:
            accumulator_scope = seq.iterator_value().scope()[-1]
        accumulator = crep.cpp_variable(unique_name("aggResult"),
                                        accumulator_scope,
                                        accumulator_type,
                                        initial_value=initial_value if initial_value is not None else crep.cpp_value(accumulator_type.default_value(), self._gc.current_scope(), accumulator_type))
        accumulator_scope.declare_variable(accumulator)

        return accumulator, accumulator_scope

    def visit_Call_Aggregate_only(self, node: ast.Call, args: List[ast.AST]):
        '''
        - (acc lambda): the accumulator is set to the first element, and the lambda is called to
                        update it after that. This is called `agg_only`.
        '''
        raise NotImplementedError()
        # This is commented out b.c. we've not written detailed testing nor found a real
        # use yet in code - though it may very well happen! So we'll leave our first guess in here.
        # agg_lambda = node.args[0]

        # # Get the sequence we are calling against and the accumulator
        # if not isinstance(node.func, ast.Attribute):
        #     raise Exception("Wrong type of function")
        # seq = self.as_sequence(node.func.value)
        # accumulator, accumulator_scope = self._create_accumulator(seq)

        # # We have to do a simple if statement here so that the first time through we can set the
        # # accumulator, and the second time we can add to it.

        # is_first_iter = crep.cpp_variable(unique_name("is_first"), self._gc.current_scope(), cpp_type=ctyp.terminal('bool'), initial_value=crep.cpp_value('true', self._gc.current_scope(), ctyp.terminal('bool')))
        # accumulator_scope.declare_variable(is_first_iter)

        # # Set the scope where we will be doing the accumulation
        # sv = seq.sequence_value()
        # if isinstance(sv, crep.cpp_sequence):
        #     self._gc.set_scope(sv.iterator_value().scope()[-1])
        # else:
        #     self._gc.set_scope(sv.scope())

        # # Code up if statement to select out the first element.
        # if_first = statement.iftest(is_first_iter)
        # self._gc.add_statement(if_first)
        # self._gc.add_statement(statement.set_var(is_first_iter, crep.cpp_value("false", self._gc.current_scope(), ctyp.terminal('bool'))))

        # # Set the accumulator
        # self._gc.add_statement(statement.set_var(accumulator, seq.sequence_value()))
        # self._gc.pop_scope()

        # # Now do the if statement and make the call to calculate the accumulation.
        # self._gc.add_statement(statement.elsephrase())
        # call = ast.Call(func=agg_lambda, args=[accumulator.as_ast(), seq.sequence_value().as_ast()])
        # self._gc.add_statement(statement.set_var(accumulator, self.get_rep(call)))

        # # Finally, since this is a terminal, we need to pop off the top.
        # self._gc.set_scope(accumulator_scope)

        # # Cache the results in our result in case we are skipping nodes in the AST.
        # node.rep = accumulator
        # self._result = accumulator

    def visit_call_Aggregate_initial(self, node: ast.Call, args: List[ast.AST]):
        '''
        - (const, acc lambda): the accumulator is set to the value, and then the lambda is called to
                        update it on every single element. This is called `agg_initial`
        '''
        raw_seq = node.args[0]
        init_val = self.get_rep(node.args[1])
        agg_lambda = node.args[2]
        assert isinstance(agg_lambda, ast.Lambda)

        # Get the sequence we are calling against and the accumulator
        seq = self.as_sequence(raw_seq)
        accumulator, accumulator_scope = self._create_accumulator(seq, initial_value=init_val, acc_type=init_val.cpp_type())

        # Now do the accumulation. This happens at the current iterator scope.
        sv = seq.sequence_value()
        if isinstance(sv, crep.cpp_sequence):
            self._gc.set_scope(sv.iterator_value().scope()[-1])
        else:
            self._gc.set_scope(sv.scope())
        call = ast.Call(func=agg_lambda, args=[accumulator.as_ast(), seq.sequence_value().as_ast()])
        update_lambda = self.get_rep(call)

        # Check the accumulator value still hols out. Since we need the accumulator previously,
        # this will allow us to patch things up. This isn't perfect, but it will do.
        if update_lambda.cpp_type().type != init_val.cpp_type().type:
            best_type = most_accurate_type([init_val.cpp_type(), update_lambda.cpp_type()])
            accumulator.update_type(best_type)

        self._gc.add_statement(statement.set_var(accumulator, update_lambda))

        # Finally, since this is a terminal, we need to pop off the top.
        self._gc.set_scope(accumulator_scope)

        # Cache the results in our result in case we are skipping nodes in the AST.
        node.rep = accumulator  # type: ignore
        self._result = accumulator

    def visit_call_Aggregate_initial_func(self, node: ast.Call, args: List[ast.AST]):
        '''
        - (start lambda, acc lambda): the accumulator is set to the start lambda call on the first
                        element in the sequence, and then acc is called to update it after that.
                        This is called `agg_initial_func`
        '''
        raise NotImplementedError()
        # Needs testing!
        # agg_lambda = node.args[1]
        # init_lambda = node.args[0]

        # # Get the sequence we are calling against and the accumulator
        # seq = self.as_sequence(node.func.value)
        # accumulator, accumulator_scope = self._create_accumulator(seq, initial_value=init_val)

        # is_first_iter = crep.cpp_value(unique_name("is_first"), accumulator_scope, cpp_type=ctyp.terminal("bool"), initial_value='true')
        # accumulator_scope.declare_variable(is_first_iter)

        # # BELOW HERE NOT CONVERTED YET.

        # # We have to initialized the variable to some value, and it depends on how the user
        # # is trying to initialize things - first iteration or with a value. We've done the value case above.
        # is_first_iter = None
        # if use_first_element_separately:
        #     is_first_iter = cpp_variable(unique_name("is_first"), self._gc.current_scope(), cpp_type="bool", initial_value='true')
        #     decl_block.declare_variable(is_first_iter)

        # # Now we need to emit code at the accumulator level.
        # self._gc.set_scope(c_loop.scope())

        # # If we have to use the first lambda to set the first value, then we need that code up front.
        # if use_first_element_separately:
        #     if_first = statement.iftest(cpp_constant(is_first_iter.as_cpp()))
        #     self._gc.add_statement(if_first)
        #     self._gc.add_statement(statement.set_var(is_first_iter, cpp_constant("false")))
        #     first_scope = self._gc.current_scope()

        #     if init_lambda is not None:
        #         call = ast.Call(init_lambda, [c_iter.as_ast()])
        #         self._gc.add_statement(statement.set_var(result, self.get_rep(call)))
        #     else:
        #         self._gc.add_statement(statement.set_var(result, c_iter))

        #     self._gc.set_scope(first_scope)
        #     self._gc.pop_scope()
        #     self._gc.add_statement(statement.elsephrase())

        # # Perform the aggregation function. We need to call it with the value and the accumulator.
        # call = ast.Call(func=agg_lambda, args=[result.as_ast(), c_iter.as_ast()])
        # self._gc.add_statement(statement.set_var(result, self.get_rep(call)))

        # # Finally, since this is a terminal, we need to pop off the top.
        # self._gc.set_scope(decl_block_scope)

        # # Cache the results in our result incase we are skipping nodes in the AST.
        # node.rep = result
        # self._result = result

    def call_Aggregate(self, node: ast.Call, args: List[ast.AST]):
        r'''Implement the aggregate algorithm in C++

        Our source we loop over, and we count out everything. The final result is whatever it is
        we are counting.

        Possible arguments to the call:

        - (acc lambda): the accumulator is set to the first element, and the lambda is called to
                        update it after that. This is called `agg_only`.
        - (const, acc lambda): the accumulator is set to the value, and then the lambda is called to
                        update it on every single element. This is called `agg_initial`
        - (start lambda, acc lambda): the accumulator is set to the start lambda call on the first
                        element in the sequence, and then acc is called to update it after that.
                        This is called `agg_initial_func`

        Limitations: only floats for now!
        '''
        # figure out which version of Aggregate we have here.
        if len(node.args) == 2:
            return self.visit_Call_Aggregate_only(node, args)
        elif len(node.args) == 3:
            if isinstance(node.args[0], ast.Lambda):
                return self.visit_call_Aggregate_initial_func(node, args)
            else:
                return self.visit_call_Aggregate_initial(node, args)

        # This isn't good!
        raise Exception("Unknown call to Aggregate. Must be Aggregate(func), Aggregate(const, func), or Aggregate(func, func)")

    def visit_Call_Member(self, call_node):
        'Method call on an object'

        # Visit everything down a level.
        self.generic_visit(call_node)

        # figure out what we are calling against, and the
        # method name we are going to be calling against.
        calling_against = self.get_rep(call_node.func.value)
        function_name = call_node.func.attr
        if not isinstance(calling_against, crep.cpp_value):
            # We didn't use get_rep_value above because now we can make a better error message.
            raise Exception("Do not know how to call '{0}' on '{1}'".format(function_name, type(calling_against).__name__))

        # We support member calls that directly translate only. Here, for example, this is only for
        # obj.pt() or similar. The translation is direct.
        c_stub = calling_against.as_cpp() + ("->" if calling_against.is_pointer() else ".")
        result_type = determine_type_mf(calling_against.cpp_type(), function_name)

        args = call_node.args
        self._result = crep.cpp_value(c_stub + function_name + f"({','.join(self.get_rep(arg).as_cpp() for arg in args)})", calling_against.scope(), result_type)

    def visit_function_ast(self, call_node):
        'Drop-in replacement for a function'
        # Get the arguments
        cpp_func = call_node.func
        arg_reps = [self.get_rep_value(a) for a in call_node.args]

        # Code up a call
        r = crep.cpp_value('{0}({1})'.format(cpp_func.cpp_name, ','.join(a.as_cpp() for a in arg_reps)), self._gc.current_scope(), cpp_type=cpp_func.cpp_return_type)

        # Include files and return the resulting expression
        for i in cpp_func.include_files:
            self._gc.add_include(i)
        call_node.rep = r
        return r

    def call_EventDataset(self, node: ast.Call, args: List[ast.AST]):
        'This has already been resolved, so return it.'
        assert hasattr(node, 'rep')
        self._result = node.rep  # type: ignore
        return node.rep  # type: ignore

    def visit_Call(self, call_node: ast.Call):
        r'''
        Very limited call forwarding.
        '''
        # What kind of a call is this?
        if isinstance(call_node.func, ast.Lambda):
            self.visit_Call_Lambda(call_node)
        elif isinstance(call_node.func, ast.Attribute):
            self.visit_Call_Member(call_node)
        elif isinstance(call_node.func, cpp_ast.CPPCodeValue):
            self._result = cpp_ast.process_ast_node(self, self._gc, call_node)
        elif isinstance(call_node.func, FunctionAST):
            self._result = self.visit_function_ast(call_node)
        else:
            # Perhaps a method call we can normalize?
            r = FuncADLNodeVisitor.visit_Call(self, call_node)
            if r is None and not hasattr(call_node, 'rep'):
                raise Exception("Do not know how to call '{0}'".format(ast.dump(call_node.func, annotate_fields=False)))
            if r is not None:
                self._result = r
        call_node.rep = self._result  # type: ignore

    def visit_Name(self, name_node: ast.Name):
        'Visiting a name - which should represent something'
        id = self.resolve_id(name_node.id)
        if isinstance(id, ast.AST):
            name_node.rep = self.get_rep(id)  # type: ignore

    def visit_Subscript(self, node):
        'Index into an array. Check types, as tuple indexing can be very bad for us'
        v = self.get_rep(node.value)
        if not isinstance(v, crep.cpp_collection):
            raise Exception("Do not know how to take the index of type '{0}'".format(v.cpp_type()))

        index = self.get_rep(node.slice)
        node.rep = crep.cpp_value("{0}.at({1})".format(v.as_cpp(), index.as_cpp()), self._gc.current_scope(), cpp_type=v.get_element_type())  # type: ignore
        self._result = node.rep

    def visit_Index(self, node):
        'We can only do single items, we cannot do slices yet'
        v = self.get_rep(node.value)
        node.rep = v
        self._result = node

    def visit_Tuple(self, tuple_node):
        r'''
        Process a tuple. We visit each component of it, and build up a representation from each result.

        See github bug #21 for the special case of dealing with (x1, x2, x3)[0].
        '''
        tuple_node.rep = crep.cpp_tuple(tuple(self.get_rep(e, retain_scope=True) for e in tuple_node.elts), self._gc.current_scope())
        self._result = tuple_node.rep

    def visit_Dict(self, node: ast.Dict):
        '''Process a dictionary. We create a C++ representation of this. The
        dict, as things currently stand, will not make its way into C++, but it might
        help us sort out some sort of output or similar.

        Args:
            node (ast.Dict): The dictionary node for us to process
        '''
        if not all(v is not None for v in node.keys):
            raise ValueError('The python construction of adding a dictionary into another dictionary is not supported ({1: 10, **old_dict})')

        values = {k: self.get_rep(v, retain_scope=True) for k, v in zip(node.keys, node.values)}
        node.rep = crep.cpp_dict(values, self._gc.current_scope())  # type: ignore
        self._result = node.rep  # type: ignore

    def visit_List(self, list_node):
        r'''
        Process a list. We visit each component of it, and build up a representation from each result.

        See github bug #21 for the special case of dealing with (x1, x2, x3)[0].
        '''
        list_node.rep = crep.cpp_tuple(tuple(self.get_rep(e, retain_scope=True) for e in list_node.elts), self._gc.current_scope())
        self._result = list_node.rep

    def visit_BinOp(self, node):
        'An in-line add'
        if type(node.op) not in _known_binary_operators:
            raise Exception(f"Do not know how to translate Binary operator {ast.dump(node.op)}!")
        left = self.get_rep(node.left)
        right = self.get_rep(node.right)

        best_type = most_accurate_type([left.cpp_type(), right.cpp_type()])
        if type(node.op) is ast.Div:
            best_type = ctyp.terminal('double', False)

        s = deepest_scope(left, right).scope()
        r = crep.cpp_value(f"({left.as_cpp()}{_known_binary_operators[type(node.op)]}{right.as_cpp()})",
                           s, best_type)

        # Cache the result to push it back further up.
        node.rep = r
        self._result = r

    def visit_UnaryOp(self, node: ast.UnaryOp):
        if type(node.op) not in _known_unary_operators:
            raise Exception(f"Do not know how to translate Unary operator {ast.dump(node.op)}!")

        operand = self.get_rep(node.operand)

        s = operand.scope()
        r = crep.cpp_value(f"({_known_unary_operators[type(node.op)]}({operand.as_cpp()}))",
                           s, operand.cpp_type())
        node.rep = r  # type: ignore
        self._result = r

    def visit_IfExp(self, node):
        r'''
        We'd like to be able to use the "?" operator in C++, but the
        problem is lazy evaluation. It could be when we look at one or the
        other item, a bunch of prep work has to be done - and that will
        show up in separate statements. So we have to use if/then/else with
        a result value.
        '''

        # The result we'll store everything in.
        result = crep.cpp_variable(unique_name("if_else_result"), self._gc.current_scope(), cpp_type=ctyp.terminal("double"))
        self._gc.declare_variable(result)

        # We always have to evaluate the test.
        current_scope = self._gc.current_scope()
        test_expr = self.get_rep(node.test)
        self._gc.add_statement(statement.iftest(test_expr))
        if_scope = self._gc.current_scope()

        # Next, we do the true and false if statement.
        self._gc.add_statement(statement.set_var(result, self.get_rep(node.body)))
        self._gc.set_scope(if_scope)
        self._gc.pop_scope()
        self._gc.add_statement(statement.elsephrase())
        self._gc.add_statement(statement.set_var(result, self.get_rep(node.orelse)))
        self._gc.set_scope(current_scope)

        # Done, the result is the rep of this node!
        node.rep = result
        self._result = result

    def visit_Compare(self, node):
        'A compare between two things. Python supports more than that, but not implemented yet.'
        if len(node.ops) != 1:
            raise Exception("Do not support 1 < a < 10 comparisons yet!")

        left = self.get_rep(node.left)
        right = self.get_rep(node.comparators[0])

        r = crep.cpp_value('({0}{1}{2})'.format(left.as_cpp(), compare_operations[type(node.ops[0])], right.as_cpp()), self._gc.current_scope(), ctyp.terminal("bool"))
        node.rep = r
        self._result = r

    def visit_BoolOp(self, node):
        '''A bool op like And or Or on a set of values
        This is a bit more complex than just "anding" things as we want to make sure to short-circuit the
        evaluation if we need to.
        '''

        # The result of this test
        result = crep.cpp_variable(unique_name('bool_op'), self._gc.current_scope(), cpp_type='bool')
        self._gc.declare_variable(result)

        # How we check and short-circuit depends on if we are doing and or or.
        check_expr = result.as_cpp() if type(node.op) == ast.And else '!{0}'.format(result.as_cpp())
        check = crep.cpp_value(check_expr, self._gc.current_scope(), cpp_type=ctyp.terminal('bool'))

        first = True
        scope = self._gc.current_scope()
        for v in node.values:
            if not first:
                self._gc.add_statement(statement.iftest(check))

            rep_v = self.get_rep(v)
            self._gc.add_statement(statement.set_var(result, rep_v))

            if not first:
                self._gc.set_scope(scope)
            first = False

        # Cache result variable so those above us have something to use.
        self._result = result
        node.rep = result

    def visit_Num(self, node):
        node.rep = crep.cpp_value(node.n, self._gc.current_scope(), guess_type_from_number(node.n))
        self._result = node.rep

    def visit_Str(self, node):
        node.rep = crep.cpp_value('"{0}"'.format(node.s), self._gc.current_scope(), ctyp.terminal("string"))
        self._result = node.rep

    def code_fill_ttree(self, e_rep: crep.cpp_rep_base, e_name: crep.cpp_variable,
                        scope_fill: Union[gc_scope, gc_scope_top_level]) -> Union[gc_scope, gc_scope_top_level]:
        '''
        How we code up setting the variables that will get collected by the 'TTree::Fill' method is a bit tricky.

        - If we have a sequence then we have to make sure to use push_back into a vector
        - If we have sequences of sequences, then we have to do multiple vector decl and push-backs, including
          declaring some extra variables
        - If the variable is set at a very low level, we need to make sure the Fill is triggered at the proper
          depth.

        Arguments:
            e_rep           XXX
            e_name          The variable that we are saving everything to
            scope_fill      The scope at which the current fill statement is going to be run. We should
                            put the statements that set the variable for collection by Fill at that scope level.

        Returns
            scope_fill      Possibly updated fill scope setting - if we were forced to go down a level (or so).

        '''
        def set_scope(scope: Union[gc_scope_top_level, gc_scope], fill_scope: Union[gc_scope, gc_scope_top_level]):
            if scope.starts_with(fill_scope):
                self._gc.set_scope(scope)
            else:
                self._gc.set_scope(fill_scope)

        # If this is a sequence of a sequence (or deeper) then we need to setup the proper variables.
        if rep_is_collection(e_rep):
            assert isinstance(e_rep, crep.cpp_sequence), \
                f'Do not know how to loop over a {type(e_rep)}'

            def fill_collection_levels(seq: crep.cpp_sequence, accumulator: crep.cpp_value):
                inner = seq.sequence_value()
                scope = seq.scope()
                if isinstance(inner, crep.cpp_sequence):
                    scope = seq.iterator_value().scope()
                    storage = crep.cpp_variable(unique_name('ntuple'), scope, cpp_type=inner.cpp_type())
                    assert not isinstance(scope, gc_scope_top_level)
                    scope.declare_variable(storage)
                    fill_collection_levels(inner, storage)
                    inner = storage

                set_scope(scope, scope_fill)
                self._gc.add_statement(statement.push_back(accumulator, inner))

            fill_collection_levels(e_rep, e_name)

        else:
            # Set the scope. Normally we want to do it where the variable was calculated
            # (think of cases when you have to calculate something with a `push_back`),
            # but if the variable was already calculated, we want to make sure we are at least
            # in the same scope as the tree fill.
            assert isinstance(e_rep, crep.cpp_value)
            set_scope(e_rep.scope(), scope_fill)

            # If the variable is something we are iterating over, then fill it, otherwise,
            # just set it.
            # if rep_is_collection(e_rep):
            #     self._gc.add_statement(statement.push_back(e_name, e_rep.sequence_value()))
            # else:
            self._gc.add_statement(statement.set_var(e_name, e_rep))
            cs = self._gc.current_scope()
            if cs.starts_with(scope_fill):
                scope_fill = cs

        return scope_fill

    def call_ResultTTree(self, node: ast.Call, args: List[ast.AST]):
        '''This AST means we are taking an iterable and converting it to a ROOT file.
        '''
        # Unpack the variables.
        assert len(args) == 4
        source = args[0]
        column_names = _extract_column_names(args[1])
        tree_name = ast.literal_eval(args[2])
        assert isinstance(tree_name, str)
        # root_filename = args[3]

        # Get the representations for each variable. We expect some sort of structure
        # for the variables - or perhaps a single variable.
        self.generic_visit(source)
        v_rep_not_norm = self.as_sequence(source)

        # What we have is a sequence of the data values we want to fill. The iterator at play
        # here is the scope we want to use to run our Fill() calls to the TTree.
        scope_fill = v_rep_not_norm.iterator_value().scope()

        # Clean the data up so it is uniform and the next bit can proceed smoothly.
        # If we don't have a tuple of data to log, turn it into a tuple.
        seq_values = v_rep_not_norm.sequence_value()
        if not isinstance(seq_values, crep.cpp_tuple):
            seq_values = crep.cpp_tuple((v_rep_not_norm.sequence_value(),), scope_fill)

        # Make sure the number of items is the same as the number of columns specified.
        if len(seq_values.values()) != len(column_names):
            raise Exception("Number of columns ({0}) is not the same as labels ({1}) in TTree creation".format(len(seq_values.values()), len(column_names)))

        # Next, look at each on in turn to decide if it is a vector or a simple variable.
        # Create a variable that we will fill for each one.
        var_names = [(name, crep.cpp_variable(unique_name(name, is_class_var=True), self._gc.current_scope(), cpp_type=get_ttree_type(rep)))
                     for name, rep in zip(column_names, seq_values.values())]

        # For each incoming variable, we need to declare something we are going to write.
        for cv in var_names:
            self._gc.declare_class_variable(cv[1])

        # Next, emit the booking code
        self._gc.add_book_statement(self.create_book_ttree_obj(tree_name, var_names))

        # Note that the output file and tree are what we are going to return.
        # The output filename is fixed - the hose code in AnalysisBase has that hard coded.
        # To allow it to be different we have to modify that template too, and pass the
        # information there. If more than one tree is written, the current code would
        # lead to a bug.
        node.rep = rh.cpp_ttree_rep("ANALYSIS.root", tree_name, self._gc.current_scope())  # type: ignore

        # For each varable we need to save, cache it or push it back, depending.
        # Make sure that it happens at the proper scope, where what we are after is defined!
        s_orig = self._gc.current_scope()
        for e_rep, e_name in zip(seq_values.values(), var_names):
            scope_fill = self.code_fill_ttree(e_rep, e_name[1], scope_fill)

        # The fill statement. This should happen at the scope where the tuple was defined.
        # The scope where this should be done is a bit tricky (note the update above):
        # - If a sequence, you want it where the sequence iterator is defined - or outside that scope
        # - If a value, you want it at the level where the value is set.
        self._gc.set_scope(scope_fill)
        self._gc.add_statement(self.create_ttree_fill_obj(tree_name))
        for e in zip(seq_values.values(), var_names):
            if rep_is_collection(e[0]):
                self._gc.add_statement(statement.container_clear(e[1][1]))

        # And we are a terminal, so pop off the block.
        self._gc.set_scope(s_orig)
        self._gc.pop_scope()
        return node.rep  # type: ignore

    def call_Select(self, node: ast.Call, args: List[ast.arg]):
        'Transform the iterable from one form to another'

        assert len(args) == 2
        source = args[0]
        selection = cast(ast.Lambda, args[1])

        # Make sure we are in a loop
        seq = self.as_sequence(source)

        # Simulate this as a "call"
        selection = lambda_unwrap(selection)
        c = ast.Call(func=selection, args=[seq.sequence_value().as_ast()])
        new_sequence_value = self.get_rep(c)

        # We need to build a new sequence.
        rep = crep.cpp_sequence(new_sequence_value, seq.iterator_value(), self._gc.current_scope())

        node.rep = rep  # type: ignore
        self._result = rep
        return rep

    def call_SelectMany(self, node: ast.AST, args: List[ast.AST]):
        r'''
        Apply the selection function to the base to generate a collection, and then
        loop over that collection.
        '''
        assert len(args) == 2
        source = args[0]
        selection = args[1]
        assert isinstance(selection, ast.Lambda)

        # Make sure the source is around. We have to do this because code generation in this
        # framework is lazy. And if the `selection` function does not use the source, and
        # looking at that source might generate a loop, that loop won't be generated! Ops!
        _ = self.as_sequence(source)

        # We need to "call" the source with the function. So build up a new
        # call, and then visit it.

        c = ast.Call(func=lambda_unwrap(selection), args=[source])

        # Get the collection, and then generate the loop over it.
        # It could be that this comes back from something that is already iterating (like a Select statement),
        # in which case we are already looping.
        seq = self.as_sequence(c)

        node.rep = seq  # type: ignore
        self._result = seq
        return seq

    def call_Where(self, node: ast.AST, args: List[ast.AST]):
        'Apply a filtering to the current loop.'

        assert len(args) == 2
        source = args[0]
        filter = args[1]
        assert isinstance(filter, ast.Lambda)

        # Make sure we are in a loop
        seq = self.as_sequence(source)

        # Simulate the filtering call - we want the resulting value to test.
        filter = lambda_unwrap(filter)
        c = ast.Call(func=filter, args=[seq.sequence_value().as_ast()])
        rep = self.get_rep(c)

        # Create an if statement
        self._gc.add_statement(statement.iftest(rep))

        # Ok - new sequence. This the same as the old sequence, only the sequence value is updated.
        # Protect against sequence of sequences (LOVE type checkers, which caught this as a possibility)
        w_val = seq.sequence_value()
        if isinstance(w_val, crep.cpp_sequence):
            raise Exception("Error: A Where clause must evaluate to a value, not a sequence")
        new_sequence_var = w_val.copy_with_new_scope(self._gc.current_scope())
        node.rep = crep.cpp_sequence(new_sequence_var, seq.iterator_value(), self._gc.current_scope())  # type: ignore

        self._result = node.rep  # type: ignore

    def call_Range(self, node: ast.AST, args: List[ast.AST]):
        'Create a collection of numbers from lower_bound'

        if len(args) != 2:
            raise Exception('blah blah')
        lower_bound = args[0]
        upper_bound = args[1]

        self._gc.add_statement(statement.block())

        element_type = ctyp.terminal("int")
        begin_value = crep.cpp_variable(unique_name("begin"), self._gc.current_scope(), element_type, initial_value=self.get_rep(lower_bound))  # type: ignore
        end_value = crep.cpp_variable(unique_name("end"), self._gc.current_scope(), element_type, initial_value=self.get_rep(upper_bound))  # type: ignore
        self._gc.declare_variable(begin_value)
        self._gc.declare_variable(end_value)

        vector_value = crep.cpp_collection(unique_name("r_obj"), self._gc.current_scope(), ctyp.collection(element_type))
        self._gc.declare_variable(crep.cpp_variable(vector_value.as_cpp(), self._gc.current_scope(),
                                                    ctyp.collection(element_type),
                                                    crep.cpp_value(f"{end_value.as_cpp()} - {begin_value.as_cpp()}",
                                                                   self._gc.current_scope(),
                                                                   ctyp.terminal("int"))))

        vector_value_begin = crep.cpp_value(f"{vector_value.as_cpp()}.begin()",
                                            self._gc.current_scope(),
                                            ctyp.terminal(f"std::vector<{element_type}>::iterator"))
        vector_value_end = crep.cpp_value(f"{vector_value.as_cpp()}.end()",
                                          self._gc.current_scope(),
                                          ctyp.terminal(f"std::vector<{element_type}>::iterator"))

        c = ast.Call(func=FunctionAST("std::iota", ["numeric"], "void"),
                     args=[vector_value_begin.as_ast(),
                           vector_value_end.as_ast(),
                           begin_value.as_ast()])

        self._gc.add_statement(statement.arbitrary_statement(self.get_rep(c).as_cpp()))

        seq = self.make_sequence_from_collection(vector_value)
        node.rep = seq  # type: ignore
        self._result = seq
        return seq

    def call_First(self, node: ast.AST, args: List[ast.AST]) -> Any:
        'We are in a sequence. Take the first element of the sequence and use that for future things.'

        # Unpack the source here
        assert len(args) == 1
        source = args[0]

        # Make sure we are in a loop.
        seq = self.as_sequence(source)

        # The First terminal works by protecting the code with a if (first_time) {} block.
        # We need to declare the first_time variable outside the block where the thing we are
        # looping over here is defined. This is a little tricky, so we delegate to another method.
        loop_scope = seq.iterator_value().scope()
        outside_block_scope = loop_scope[-1]

        # Define the variable to track this outside that block.
        is_first = crep.cpp_variable(unique_name('is_first'),
                                     outside_block_scope,
                                     cpp_type=ctyp.terminal('bool'),
                                     initial_value=crep.cpp_value('true', self._gc.current_scope(), ctyp.terminal('bool')))
        outside_block_scope.declare_variable(is_first)

        # Now, as long as is_first is true, we can execute things inside this statement.
        # The trick is putting the if statement in the right place. We need to locate it just one level
        # below where we defined the scope above.
        s = statement.iftest(is_first)
        s.add_statement(statement.set_var(is_first, crep.cpp_value('false', top_level_scope(), cpp_type=ctyp.terminal('bool'))))

        sv = seq.sequence_value()
        if isinstance(sv, crep.cpp_sequence):
            self._gc.set_scope(sv.iterator_value().scope()[-1])
        else:
            self._gc.set_scope(sv.scope())
        self._gc.add_statement(s)

        # If we just found the first sequence in a sequence, return that.
        # Otherwise return a new version of the value.
        first_value = sv if isinstance(sv, crep.cpp_sequence) else sv.copy_with_new_scope(self._gc.current_scope())

        node.rep = first_value  # type: ignore
        self._result = first_value



# Collected code to get collections from the event object
import ast
import copy
from abc import ABC, abstractmethod

import func_adl_xAOD.common.cpp_ast as cpp_ast
import func_adl_xAOD.common.cpp_representation as crep
import func_adl_xAOD.common.cpp_types as ctyp
from func_adl_xAOD.common.cpp_vars import unique_name
from func_adl_xAOD.common.math_utils import get_math_methods


# Need a type for our type system to reason about the containers.
class event_collection_container(ABC):
    def __init__(self, type_name, is_pointer):
        self._type_name = type_name
        self._is_pointer = is_pointer

    def is_pointer(self):
        return self._is_pointer

    @abstractmethod
    def __str__(self):
        pass


class event_collection_collection(event_collection_container):
    def __init__(self, type_name, element_name, is_type_pointer, is_element_pointer):
        event_collection_container.__init__(self, type_name, is_type_pointer)
        self._element_name = element_name
        self._is_pointer = is_element_pointer

    def element_type(self):
        return ctyp.terminal(self._element_name, is_pointer=self._is_pointer)

    def dereference(self):
        'Return a new version of us that is not a pointer'
        new_us = copy.copy(self)
        new_us._is_pointer = False
        return new_us


class event_collections(ABC):
    def __init__(self, collections):
        self._collections = collections

    @abstractmethod
    def get_running_code(self, container_type: event_collection_container) -> list:
        pass

    def get_collection(self, info, call_node):
        r'''
        Return a cpp ast for accessing the jet collection
        '''
        # Get the name jet collection to look at.
        if len(call_node.args) != 1:
            raise ValueError(f"Calling {info['function_name']} - only one argument is allowed")
        if not isinstance(call_node.args[0], ast.Str):
            raise ValueError(f"Calling {info['function_name']} - only acceptable argument is a string")

        # Fill in the CPP block next.
        r = cpp_ast.CPPCodeValue()
        r.args = ['collection_name', ]
        r.include_files += info['include_files']

        r.running_code += self.get_running_code(info['container_type'])
        r.result = 'result'

        is_collection = info['is_collection'] if 'is_collection' in info else True
        if is_collection:
            r.result_rep = lambda scope: crep.cpp_collection(unique_name(info['function_name'].lower()), scope=scope, collection_type=info['container_type'])  # type: ignore
        else:
            r.result_rep = lambda scope: crep.cpp_variable(unique_name(info['function_name'].lower()), scope=scope, cpp_type=info['container_type'])

        # Replace it as the function that is going to get called.
        call_node.func = r

        return call_node

    # Config everything.
    def create_higher_order_function(self, info):
        'Creates a higher-order function because python scoping is broken'
        return lambda call_node: self.get_collection(info, call_node)

    def get_method_names(self):
        method_names = {}
        for info in self._collections:
            method_names[info['function_name']] = self.create_higher_order_function(info)
        method_names.update(get_math_methods())
        return method_names



from func_adl_xAOD.cms.aod.event_collections import \
    cms_aod_event_collections as ec
from func_adl_xAOD.cms.aod.query_ast_visitor import cms_aod_query_ast_visitor
from func_adl_xAOD.common.executor import executor


class cms_aod_executor(executor):
    def __init__(self):
        file_names = ['analyzer_cfg.py', 'Analyzer.cc', 'BuildFile.xml', 'runner.sh']
        runner_name = 'runner.sh'
        template_dir_name = 'func_adl_xAOD/template/cms/r5'
        super().__init__(file_names, runner_name, template_dir_name, ec().get_method_names())

    def get_visitor_obj(self):
        return cms_aod_query_ast_visitor()



# Code to aid with accessing jet collections
import ast

import func_adl_xAOD.common.cpp_ast as cpp_ast
import func_adl_xAOD.common.cpp_representation as crep
import func_adl_xAOD.common.cpp_types as ctyp
from func_adl_xAOD.common.cpp_vars import unique_name


def getAttributeFloatAst(call_node: ast.Call):
    r'''
    Return an attribute on one of the xAOD objects.
    '''
    # Get the name of the moment out
    if len(call_node.args) != 1:
        raise Exception("Calling getAttributeFloat - only one argument is allowed")
    if not isinstance(call_node.args[0], ast.Str):
        raise Exception("Calling getAttributeFloat - only acceptable argument is a string")

    r = cpp_ast.CPPCodeValue()
    # We don't need include files for this - just quick access
    r.args = ['moment_name', ]
    r.replacement_instance_obj = ('obj_j', call_node.func.value.id)  # type: ignore
    r.running_code += ['float result = obj_j->getAttribute<float>(moment_name);']
    r.result = 'result'
    r.result_rep = lambda sc: crep.cpp_variable(unique_name("jet_attrib"), scope=sc, cpp_type=ctyp.terminal('float'))

    # Replace it as the function that is going to get called.
    call_node.func = r  # type: ignore

    return call_node


def getAttributeVectorFloatAst(call_node: ast.Call):
    r'''
    Return a cpp ast accessing a vector of doubles for an xAOD attribute
    '''
    # Get the name of the moment out
    if len(call_node.args) != 1:
        raise Exception("Calling getAttributeVectorFloat - only one argument is allowed")
    if not isinstance(call_node.args[0], ast.Str):
        raise Exception("Calling getAttributeVectorFloat - only acceptable argument is a string")

    r = cpp_ast.CPPCodeValue()
    r.include_files += ['vector']
    r.args = ['moment_name', ]
    r.replacement_instance_obj = ('obj_j', call_node.func.value.id)  # type: ignore
    r.running_code += ['auto result = obj_j->getAttribute<std::vector<double>>(moment_name);']
    r.result = 'result'
    r.result_rep = lambda sc: crep.cpp_collection(unique_name("jet_vec_attrib_"), scope=sc, collection_type=ctyp.collection(ctyp.terminal('double')))  # type: ignore

    # Replace it as the function that is going to get called.
    call_node.func = r  # type: ignore

    return call_node


def getAttribute(call_node: ast.Call):
    '''
    This is a common mistake as this is how it is written in C++. However, because we do
    not do templates, we need to do something more explicit. So, bomb out here.
    '''
    raise Exception("Do not call `getAttribute` - instead call `getAttributeFloat` or `getAttributeVectorFloat`.")


def get_jet_methods():
    return {
        'getAttribute': getAttribute,
        'getAttributeFloat': getAttributeFloatAst,
        'getAttributeVectorFloat': getAttributeVectorFloatAst
    }



# Statements
from abc import ABC, abstractmethod  # For declaring abstract base class
from typing import Any

import func_adl_xAOD.common.cpp_representation as crep


class BlockException (Exception):
    def __init__(self, message):
        Exception.__init__(self, message)


class block:
    'This is a block of statements surrounded by a scoping (like open close bracket, for loop, etc.)'

    def __init__(self):
        self._statements = []
        self._variables = []
        self._rep_dict = {}

    def add_statement(self, s):
        'Add statement s to the list of statements'
        self._statements += [s]

    def declare_variable(self, n):
        'Declare a variable n, which is of type cpp_variable'
        self._variables += [n]

    def emit(self, e):
        'Render the block of code'
        e.add_line("{")
        for v in self._variables:
            init_value = "" if not isinstance(v, crep.cpp_variable) or not v.initial_value() else " ({0})".format(v.initial_value().as_cpp())
            e.add_line("{0} {1}{2};".format(v.cpp_type(), v.as_cpp(), init_value))
        for s in self._statements:
            s.emit(e)
        e.add_line("}")

    def get_rep(self, name: Any) -> Any:
        '''Return the representation for some object. If we do not know its value
        then ask our parent for the value. Return None if we can't find it.

        Args:
            name:           Key for lookup

        Returns:
            None if there is nothing defined with that key in the hierarchy, or
            an actual value if there is.
        '''
        if name in self._rep_dict:
            return self._rep_dict[name]
        return None

    def set_rep(self, name: Any, value: Any):
        '''Defines the `value` for a lookup of `name` at any time in this block and
        below.

        Args:
            name:       The lookup key
            value       The value to be cached
        '''
        if name in self._rep_dict:
            raise BlockException(f'Internal Error: Representation for {str(name)} already exists. Cannot set twice')
        self._rep_dict[name] = value


class loop(block):
    'A for loop'

    def __init__(self, loop_var_rep: crep.cpp_value,
                 collection_rep: crep.cpp_collection,
                 is_loop_var_a_pntr=False, is_loop_var_a_ref=False):
        '''
        Create a new implicit for loop statement. A new var is created, and the scope is set to
        be the one down from here.
        '''
        block.__init__(self)
        self._collection = collection_rep
        self._loop_variable = loop_var_rep
        self._is_loop_var_a_pointer = is_loop_var_a_pntr
        self._is_loop_var_a_reference = is_loop_var_a_ref

    def emit(self, e):
        'Emit a for loop enclosed by a block of code'
        e.add_line("for (auto {0}{1} : {2})".format(
            '*' if self._is_loop_var_a_pointer
            else ('&' if self._is_loop_var_a_reference else ''),
            self._loop_variable.as_cpp(), self._collection.as_cpp()))
        block.emit(self, e)


class iftest(block):
    'An if statement'

    def __init__(self, if_expr):
        block.__init__(self)
        self._expr = if_expr

    def emit(self, e):
        e.add_line('if ({0})'.format(self._expr.as_cpp()))
        block.emit(self, e)


class elsephrase(block):
    'An else statement. Must come after you pop and if statement off'

    def __init__(self):
        block.__init__(self)

    def emit(self, e):
        'Emit an else statement'
        e.add_line('else')
        block.emit(self, e)


# By Inheriting from ABC we declare it as an Abstract Base Class
class book_ttree(ABC):
    'Book a TTree for writing out. Meant to be in the Book method'

    def __init__(self, tree_name, leaves):
        self._tree_name = tree_name
        self._leaves = leaves

    @abstractmethod
    def emit(self, e):
        # It is marked as an abstract method and will have to be
        # implemented before being used
        pass


class ttree_fill(ABC):
    'Fill a TTree'

    def __init__(self, tree_name):
        self._tree_name = tree_name

    @abstractmethod
    def emit(self, e):
        pass


class set_var:
    'Assing a value to a variable'

    def __init__(self, target_var, value_var):
        r'''
        target_var, value_var: representations we will use
        '''
        self._target = target_var
        self._value = value_var

    def emit(self, e):
        e.add_line('{0} = {1};'.format(self._target.as_cpp(), self._value.as_cpp()))


class push_back:
    'push a variable onto a vector'

    def __init__(self, target_collection, value_var):
        r'''
        target_col, value_var: representations we will use
        '''
        self._target = target_collection
        self._value = value_var

    def emit(self, e):
        e.add_line('{0}.push_back({1});'.format(self._target.as_cpp(), self._value.as_cpp()))


class container_clear:
    'push a variable onto a vector'

    def __init__(self, collection):
        r'''
        target_col, value_var: representations we will use
        '''
        self._collection = collection

    def emit(self, e):
        e.add_line('{0}.clear();'.format(self._collection.as_cpp()))


class arbitrary_statement:
    'An arbitrary line of C++ code. Avoid if possible, as it makes analysis impossible'

    def __init__(self, line: str):
        self._line = line

    def emit(self, e):
        ll = self._line
        if not ll.endswith(';'):
            ll += ';'
        e.add_line(ll)



import func_adl_xAOD.common.cpp_types as ctyp
from func_adl_xAOD.common.event_collections import (
    event_collection_collection, event_collection_container, event_collections)


class cms_aod_event_collection_container(event_collection_container):
    def __init__(self, type_name, is_pointer=True):
        super().__init__(type_name, is_pointer)

    def __str__(self):
        return f"edm::Handle<{self._type_name}>"


class cms_aod_event_collection_collection(event_collection_collection):
    def __init__(self, type_name, element_name, is_type_pointer=True, is_element_pointer=True):
        super().__init__(type_name, element_name, is_type_pointer, is_element_pointer)

    def __str__(self):
        return f"edm::Handle<{self._type_name}>"


# all the collections types that are available. This is required because C++
# is strongly typed, and thus we have to transmit this information.
cms_aod_collections = [
    {
        'function_name': "Tracks",
        'include_files': ['DataFormats/TrackReco/interface/Track.h', 'DataFormats/TrackReco/interface/TrackFwd.h'],
        'container_type': cms_aod_event_collection_collection('reco::TrackCollection', 'reco::Track')
    },
    {
        'function_name': "TrackMuons",
        'include_files': ['DataFormats/MuonReco/interface/Muon.h'],
        'container_type': cms_aod_event_collection_collection('reco::TrackCollection', 'reco::Track')
    },
    {
        'function_name': "Muons",
        'include_files': ['DataFormats/MuonReco/interface/Muon.h'],
        'container_type': cms_aod_event_collection_collection('reco::MuonCollection', 'reco::Muon')
    },
]


class cms_aod_event_collections(event_collections):
    def __init__(self):
        super().__init__(cms_aod_collections)

    def get_running_code(self, container_type: event_collection_container) -> list:
        return [f'{container_type} result;',
                'iEvent.getByLabel(collection_name, result);']


ctyp.add_method_type_info("reco::Track", "hitPattern", ctyp.terminal('reco::HitPattern'))
ctyp.add_method_type_info("reco::Muon", "hitPattern", ctyp.terminal('reco::HitPattern'))



#
# Read the submission directory as a command line argument. You can
# extend the list of arguments with your private ones later on.
# Set up (Py)ROOT.
import ROOT  # type: ignore
import optparse
from AnaAlgorithm.DualUseConfig import createAlgorithm  # type: ignore

parser = optparse.OptionParser()

ROOT.xAOD.Init().ignore()
parser.add_option('-s', '--submission-dir', dest='submission_dir',
                  action='store', type='string', default='submitDir',
                  help='Submission directory for EventLoop')
(options, args) = parser.parse_args()


# Set up the sample handler object. See comments from the C++ macro
# for the details about these lines.
sh = ROOT.SH.SampleHandler()
sh.setMetaString('nc_tree', 'CollectionTree')
# ROOT.SH.ScanDir().filePattern( '{{df}}' ).scan( sh, inputFilePath )
ROOT.SH.readFileList(sh, "ANALYSIS", "filelist.txt")
sh.printContent()

# Create an EventLoop job.
job = ROOT.EL.Job()
job.sampleHandler(sh)
# job.options().setDouble( ROOT.EL.Job.optMaxEvents, 500 )

# Commented out for now because it really slows things down. Uncomment and change
# the bank to be Analysis_NOSYS in query.cxx and it will work again.
# #Get the systematics tool in - because we need it.
# from AnaAlgorithm.AnaAlgorithmConfig import AnaAlgorithmConfig
# config = AnaAlgorithmConfig( 'CP::SysListLoaderAlg/SysLoaderAlg' )
# config.sigmaRecommended = 1
# job.algsAdd( config )

# # First step - run calibration for the jets so they are available to use when we want them.
# ROOT.CP.JetCalibrationAlg ("dummy", None)
# from JetAnalysisAlgorithms.JetAnalysisSequence import makeJetAnalysisSequence
# #from jetsequence import makeJetAnalysisSequence
# jetSequence = makeJetAnalysisSequence( 'data', "AntiKt4EMTopoJets" )
# jetSequence.configure( inputName = "AntiKt4EMTopoJets", outputName = 'AnalysisJets' )
# for alg in jetSequence:
#     job.algsAdd(alg)

# Create the algorithm's configuration.
alg = createAlgorithm('query', 'AnalysisAlg')
# later on we'll add some configuration options for our algorithm that go here

# Add our algorithm to the job
job.algsAdd(alg)
job.outputAdd(ROOT.EL.OutputStream('ANALYSIS'))

# Run the job using the direct driver.
driver = ROOT.EL.DirectDriver()
driver.submit(job, options.submission_dir)



# setuptools loads some plugins necessary for use here.
import os
from distutils.core import setup

from version_info import version_func_adl

# Use the readme as the long description.
with open("README.md", "r") as fh:
    long_description = fh.read()

# Get the version number right
version = os.getenv('func_adl_xaod_version')
if version is None:
    version = '0.0.0-alpha.10'
else:
    version = version.split('/')[-1]

# Finally, the setup.
setup(name="func_adl_xAOD",
      version=version,
      packages=['func_adl_xAOD'] + [f'func_adl_xAOD.{f}' for f in ['atlas.xaod', 'cms.aod', 'common']],
      scripts=[],
      description="Functional Analysis Description Language backend for accessing ATLAS xAOD files.",
      long_description=long_description,
      long_description_content_type="text/markdown",
      author="G. Watts (IRIS-HEP/UW Seattle)",
      author_email="gwatts@uw.edu",
      maintainer="Gordon Watts (IRIS-HEP/UW Seattle)",
      maintainer_email="gwatts@uw.edu",
      url="https://github.com/iris-hep/func_adl_xAOD",
      license="TBD",
      test_suite="tests",
      install_requires=[
          "requests~=2.0",
          "retry~=0.9",
          "jinja2",
          "qastle",
          f"func_adl.ast{version_func_adl}",
      ],
      extras_require={
          'test': [
              'pytest>=3.9',
              'pytest-asyncio',
              'pytest-mock',
              'pytest-cov',
              'coverage',
              'flake8',
              'codecov',
              'autopep8',
              'twine',
              'testfixtures',
              'wheel',
              'asyncmock',
              "pandas~=1.0",
              "uproot~=3.7",
          ],
      },
      classifiers=[
          "Development Status :: 3 - Alpha",
          # "Development Status :: 4 - Beta",
          # "Development Status :: 5 - Production/Stable",
          # "Development Status :: 6 - Mature",
          "Intended Audience :: Developers",
          "Intended Audience :: Information Technology",
          "Programming Language :: Python",
          "Topic :: Software Development",
          "Topic :: Utilities",
      ],
      package_data={
          'func_adl_xAOD': ['template/atlas/r21/*', 'template/cms/r5/*'],
      },
      platforms="Any",
      )



# Infrastructure to replace simple functions (like "abs") in python with their
# equivalent in C++.
import ast
from collections import namedtuple


class FunctionAST(ast.AST):
    '''
    An AST node that represents a function that can be a drop-in replacement for
    a python function
    '''

    def __init__(self, cpp_name, include_files, cpp_return_type):
        ''' Initialize an AST node that represents a C++ drop-in function call

        cpp_name: The C++ name that we will use to do the call
        include_files: List of include files to be put at the start of the emitted source.
        '''
        self.cpp_name = cpp_name
        self.include_files = include_files
        self.cpp_return_type = cpp_return_type
        self.fields = []


class find_known_functions(ast.NodeTransformer):
    def visit_Call(self, node):
        'Look for a call to a Name function that is in our list'
        # First go one level down.
        self.generic_visit(node)

        # See if we are candidate for replacement
        # Get a fully qualified name as possible.
        if type(node.func) is not ast.Name:
            return node

        try:
            fnc = eval(node.func.id)
            fnc_name = '{0}.{1}'.format(fnc.__module__, node.func.id)
        except NameError:
            fnc_name = node.func.id

        if fnc_name not in functions_to_replace:
            return node

        # Build the replacement.
        info = functions_to_replace[fnc_name]
        node.func = FunctionAST(info.cpp_name, info.include_files, info.cpp_return_type)

        return node


def add_function_mapping(python_name, cpp_name, include_files, return_type):
    '''Add a re-mapping from a python function to an actual function.

    python_name: fully qualified name of the python function
    cpp_name: fully qualified name of the C++ function
    include_files: any include files that should be included in the C++ source. Can also be a single string.
    return_type: C++ return type
    '''
    global functions_to_replace
    functions_to_replace[python_name] = cpp_function(cpp_name, include_files if type(include_files) is list else [include_files, ], return_type)


# The list of functions to do the replacement
functions_to_replace = {}


# The named tuple that stores the replacement info.
cpp_function = namedtuple('cpp_function', ['cpp_name', 'include_files', 'cpp_return_type'])

# CMATH functions - the catalog was pulled from
# http://www.cplusplus.com/reference/cmath/
# Trig
add_function_mapping('sin', 'std::sin', 'cmath', 'double')
add_function_mapping('cos', 'std::cos', 'cmath', 'double')
add_function_mapping('tan', 'std::tan', 'cmath', 'double')
add_function_mapping('acos', 'std::acos', 'cmath', 'double')
add_function_mapping('asin', 'std::asin', 'cmath', 'double')
add_function_mapping('atan', 'std::atan', 'cmath', 'double')
add_function_mapping('atan2', 'std::atan2', 'cmath', 'double')

# Hyperbolic
add_function_mapping('sinh', 'std::sinh', 'cmath', 'double')
add_function_mapping('cosh', 'std::cosh', 'cmath', 'double')
add_function_mapping('tanh', 'std::tanh', 'cmath', 'double')
add_function_mapping('asinh', 'std::asinh', 'cmath', 'double')
add_function_mapping('acosh', 'std::acosh', 'cmath', 'double')
add_function_mapping('atanh', 'std::atanh', 'cmath', 'double')

# Exponential and Log functions
add_function_mapping('exp', 'std::exp', 'cmath', 'double')
# add_function_mapping('frexp', 'std::frexp', 'cmath', 'double')  # By value parameter
add_function_mapping('ldexp', 'std::ldexp', 'cmath', 'double')
add_function_mapping('log', 'std::log', 'cmath', 'double')
add_function_mapping('ln', 'std::log', 'cmath', 'double')
add_function_mapping('log10', 'std::log10', 'cmath', 'double')
# add_function_mapping('modf', 'std::modf', 'cmath', 'double')  # By value parameter
add_function_mapping('exp2', 'std::exp2', 'cmath', 'double')
add_function_mapping('expm1', 'std::expm1', 'cmath', 'double')
add_function_mapping('ilogb', 'std::ilogb', 'cmath', 'double')
add_function_mapping('log1p', 'std::log1p', 'cmath', 'double')
add_function_mapping('log2', 'std::log2', 'cmath', 'double')
add_function_mapping('scalbn', 'std::scalbn', 'cmath', 'double')
add_function_mapping('scalbln', 'std::scalbln', 'cmath', 'double')

# Power Functions
add_function_mapping('pow', 'std::pow', 'cmath', 'double')
add_function_mapping('sqrt', 'std::sqrt', 'cmath', 'double')
add_function_mapping('cbrt', 'std::cbrt', 'cmath', 'double')
add_function_mapping('hypot', 'std::hypot', 'cmath', 'double')

# Error and Gamma Functions
add_function_mapping('erf', 'std::erf', 'cmath', 'double')
add_function_mapping('erfc', 'std::erfc', 'cmath', 'double')
add_function_mapping('tgamma', 'std::tgamma', 'cmath', 'double')
add_function_mapping('lgamma', 'std::lgamma', 'cmath', 'double')

# Rounding and remainder functions
add_function_mapping('ceil', 'std::ceil', 'cmath', 'double')
add_function_mapping('floor', 'std::ceil', 'cmath', 'double')
add_function_mapping('fmod', 'std::ceil', 'cmath', 'double')
add_function_mapping('trunc', 'std::ceil', 'cmath', 'double')
add_function_mapping('round', 'std::ceil', 'cmath', 'double')
# add_function_mapping('lround', 'std::ceil', 'cmath', 'double')  # Long int
# add_function_mapping('llround', 'std::ceil', 'cmath', 'double')  # Long int
add_function_mapping('rint', 'std::ceil', 'cmath', 'double')
# add_function_mapping('lrint', 'std::ceil', 'cmath', 'double')  # Long int
# add_function_mapping('llrint', 'std::ceil', 'cmath', 'double')  # Long int
add_function_mapping('nearbyint', 'std::ceil', 'cmath', 'double')
add_function_mapping('remainder', 'std::ceil', 'cmath', 'double')
add_function_mapping('remquo', 'std::ceil', 'cmath', 'double')

# Floating-point manipulation functions
add_function_mapping('copysign', 'std::copysign', 'cmath', 'double')
add_function_mapping('nan', 'std::nan', 'cmath', 'double')
add_function_mapping('nextafter', 'std::nextafter', 'cmath', 'double')
add_function_mapping('nexttoward', 'std::nexttoward', 'cmath', 'double')

# Minimum, maximum, difference functions
add_function_mapping('fdim', 'std::fdim', 'cmath', 'double')
add_function_mapping('fmax', 'std::fmax', 'cmath', 'double')
add_function_mapping('fmin', 'std::fmin', 'cmath', 'double')

# Other functions
add_function_mapping('fabs', 'std::fabs', 'cmath', 'double')
add_function_mapping('abs', 'std::fabs', 'cmath', 'double')
add_function_mapping('fma', 'std::fma', 'cmath', 'double')

# Python built-in functions. These are weird as they get a different name, and
# don't really come in as a function.
add_function_mapping('builtins.abs', 'std::abs', 'cmath', 'double')
add_function_mapping('builtins.pow', 'std::pow', 'cmath', 'double')  # Not the 3 arg version!

# add_function_mapping('builtins.max', 'std::abs', 'cmath', 'double')  # Do not use to do a sequence
# add_function_mapping('builtins.min', 'std::abs', 'cmath', 'double')  # Do not use to do a sequence
